//  BLE test code to be used on an ESP32 (peripheral) in conjunction with an AI2 app (central) 
//  release 1.0
//  date 2024/04/04
//  Based on Neil Kolban example for IDF: https://github.com/nkolban/esp32-snippets/blob/master/cpp_utils/tests/BLE%20Tests/SampleNotify.cpp
//  Ported to Arduino ESP32 by Evandro Copercini
//  Modified by Ugo Schiara

/* Creates a BLE server that, once a connection is performed, at every character received, it sends back the same character
   The service advertises itself as: 6E400001-B5A3-F393-E0A9-E50E24DCCA9E
   Has a characteristic of: 6E400002-B5A3-F393-E0A9-E50E24DCCA9E - used for receiving data with "WRITE" 
   Has a characteristic of: 6E400003-B5A3-F393-E0A9-E50E24DCCA9E - used to send data with  "NOTIFY"

   The flow to create the BLE server is:
   1. Create a BLE Server
   2. Create a BLE Service
   3. Create a BLE Characteristic on the Service
   4. Create a BLE Descriptor on the characteristic
   5. Start the service.
   6. Advertises once at every character received by sending back the same character

   In this example rxValue is the data received (only accessible inside the function) that is sent back to the Central (AI2 client)
*/
#include <BLEDevice.h>
#include <BLEServer.h>
#include <BLEUtils.h>
#include <BLE2902.h>

BLEServer* pServer = NULL;
BLECharacteristic* pTxCharacteristic;
bool deviceConnected = false;                             // flag to recognize if the client has disconnected
bool oldDeviceConnected = false;                          // and to perform a new connection attempt  


// See the following for generating UUIDs:
// https://www.uuidgenerator.net/

#define SERVICE_UUID           "6E400001-B5A3-F393-E0A9-E50E24DCCA9E"     // UART service UUID
#define CHARACTERISTIC_UUID_RX "6E400002-B5A3-F393-E0A9-E50E24DCCA9E"     // its characteristics
#define CHARACTERISTIC_UUID_TX "6E400003-B5A3-F393-E0A9-E50E24DCCA9E"


class MyServerCallbacks: public BLEServerCallbacks {
    void onConnect(BLEServer* pServer) {
      deviceConnected = true;
    };

    void onDisconnect(BLEServer* pServer) {
      deviceConnected = false;
    }
};

class MyCallbacks: public BLECharacteristicCallbacks {
    void onWrite(BLECharacteristic *pCharacteristic) {
      std::string rxValue = pCharacteristic->getValue();
      std::string myStringForUnit8((char*)&rxValue[0], 1);        // trick to allow sending back one single character to the client
                                                                  // please refer to https://github.com/nkolban/esp32-snippets/issues/79
      if (rxValue.length() > 0) {                                 // Something has been received
        Serial.println("*********");                              // starts advertising on Serial Monitor
        Serial.print("Received Value: ");
        for (int i = 0; i < rxValue.length()-1; i++)              // AI2 sends the Linefeed to close the string: it shall be discarded
                                                                  // that's why .length()-1    
          Serial.print(rxValue[i]);                               // shows the received char on Serial Monitor
        pTxCharacteristic->setValue(myStringForUnit8);            // sends back what has been received (just one char)
        pTxCharacteristic->notify();                    
        Serial.println();                                     
        Serial.println("*********");                              // ends advertising on Serial Monito
      }
    }
};


void setup() {
  Serial.begin(115200);                                           // baudrate for Serial Monitor

  // Create the BLE Device
  BLEDevice::init("UART Service");                                // BLE Server Name (to be searched for by the AI2 app)

  // Create the BLE Server
  pServer = BLEDevice::createServer();
  pServer->setCallbacks(new MyServerCallbacks());

  // Create the BLE Service
  BLEService *pService = pServer->createService(SERVICE_UUID);

  // Create a BLE Characteristic
  pTxCharacteristic = pService->createCharacteristic(
										CHARACTERISTIC_UUID_TX,
										BLECharacteristic::PROPERTY_NOTIFY                    // used to send characters to AI2 app
									);
                      
  pTxCharacteristic->addDescriptor(new BLE2902());

  BLECharacteristic *pRxCharacteristic = pService->createCharacteristic(
											 CHARACTERISTIC_UUID_RX,
											BLECharacteristic::PROPERTY_WRITE                  // used to receive characters from AI2 app
										);

  pRxCharacteristic->setCallbacks(new MyCallbacks());

  // Start the service
  pService->start();

  // Start advertising
  pServer->getAdvertising()->start();
  Serial.println("Waiting a client connection to notify...");
}

void loop() {

    if (deviceConnected)
    {
		    Delay(100);                               // just leaves some breath to the bluetooth stack 
	  }

    // disconnecting
    if (!deviceConnected && oldDeviceConnected) 
    {
        Delay(500);                               // gives the bluetooth stack the chance to get things ready
        pServer->startAdvertising();              // restarts advertising
        Serial.println("start advertising");
        oldDeviceConnected = deviceConnected;
    }
    // connecting
    if (deviceConnected && !oldDeviceConnected) {
		// do stuff here on connecting
        oldDeviceConnected = deviceConnected;
    }
}

//-------------------------------------------------------------------------
// Private delay function that does not block the CPU
// Input parameter number of millisecond to wait (unsigned long)
void Delay(unsigned long TimeToWait)
{
  unsigned long now = millis();                         // reads the clock 
  while ((millis()-now) < TimeToWait);                  // waits until the delta milliseconds is smaller than the wanted timeout
                                                        // but it does not block the CPU, thus leaving the interrupts to work
}                                                        